"""This module contains `docker image build` class"""

import sys
from io import BytesIO

from .command import Command
from ...Utils.json_iterparse import json_iterparse
from ...Utils.switch import switch
from ...Utils.urlutil import (is_git_url, is_url)

class Build(Command):
    """This class implements `docker image build` command"""

    name = "image build"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def preprocess(self, args):
        """Preprocess build arguments"""
        for case in switch(args["path"]):
            if case('-'):
                del args["path"]
                args['fileobj'] = BytesIO(sys.stdin.read())
                break
            if is_git_url(args["path"]):
                break
            if is_url(args["path"]):
                break
            if case():
                break
        args["buildargs"] = dict(args["buildargs"]) if args["buildargs"] else None

    def eval_command(self, args):
        self.preprocess(args)
        for line in self.client.build(**args):
            for iterElement in list(json_iterparse(line)):
                print iterElement["stream"],
        self.settings[self.name] = ""

    def final(self):
        return self.settings[self.name]



"""This module provides switch functionality"""

class switch(object):
    """An implementation of classical switch method"""
    def __init__(self, value):
        self.value = value
        self.fall = False

    def __iter__(self):
        """Return the match method once, then stop"""
        yield self.match
        raise StopIteration

    def match(self, *args):
        """Indicate whether or not to enter a case suite"""
        if self.fall or not args:
            return True
        elif self.value in args:
            self.fall = True
            return True
        else:
            return False



"""This module contains `docker version` class"""

from platform import python_version, system, machine
import pystache
import arrow
from .... import __version__

from .command import Command


class Version(Command):
    """This class implements `docker version` command"""

    name = "version"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        versions = self.client.version()
        versions["ClientVersion"] = __version__
        versions["PythonVersion"] = python_version()
        versions["ClientOs"] = system()
        versions["ClientArch"] = machine()
        versions["BuildTime"] = arrow.get(versions["BuildTime"]).humanize()

        self.settings[self.name] = pystache.render("""\
Client:
    Version:         {{ClientVersion}}
    Python version:  {{PythonVersion}}
    OS/Arch:         {{ClientOs}}/{{ClientArch}}

Server:
    Version:         {{Version}}
    API version:     {{ApiVersion}} (minimum version {{MinAPIVersion}})
    Go version:      {{GoVersion}}
    Git commit:      {{GitCommit}}
    Built:           {{BuildTime}}
    OS/Arch:         {{Os}}/{{Arch}}
    Kernel version:  {{KernelVersion}}
    Experimental:    {{Experimental}}
""", versions)

    def final(self):
        return self.settings[self.name]



"""This module contains `docker container create` class"""

from docker.errors import APIError

from .command import Command

SECOND = 1000000000


class Create(Command):
    """This class implements `docker container create` command"""

    name = "container create"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def preprocess(self, args):
        """Preprocess arguments provided for `docker container create`"""
        # port_bindings
        ports = []
        if args.get("port_bindings"):
            port_bindings = args["port_bindings"]
            args["port_bindings"] = {}
            for port_binding in port_bindings:
                host, container = port_binding.strip('{}').split(':', 2)
                args["port_bindings"].update([[container, host], ])
                ports.append(container)

        # binds
        volumes = []
        if args.get("binds"):
            volume_bindings = args["binds"]
            args["binds"] = []
            for volume_binding in volume_bindings:
                try:
                    host, container, _ = volume_binding.split(':', 3)
                except ValueError:
                    host, container = volume_binding.split(':', 2)
                args["binds"].append(volume_binding)
                volumes.append(container)

        # image
        image = args["image"]
        del args["image"]

        # command
        if "command" in args:
            command = args["command"]
            del args["command"]

        # hostname
        if "hostname" in args:
            hostname = args["hostname"]
            del args["hostname"]

        # cidfile
        if args.get("cidfile"):
            cidfile = args["cidfile"]
        else:
            cidfile = None
        del args["cidfile"]

        # detach
        detach = args["detach"]
        del args["detach"]

        # entrypoint
        entrypoint = args["entrypoint"]
        del args["entrypoint"]

        # environment
        environment = args["environment"]
        del args["environment"]

        # healthcheck
        have_health_settings = bool(args["health_cmd"]) | bool(args["health_interval"]) | bool(
            args["health_timeout"]) | bool(args["health_retries"])
        if args["no_healthcheck"]:
            healthcheck = None
        elif have_health_settings:
            from ...Utils.duration_conversion import from_str
            healthcheck = dict(
                test=args["health_cmd"] if args["health_cmd"] else None,
                interval=from_str(
                    args["health_interval"]).seconds * SECOND if args["health_interval"] else None,
                timeout=from_str(args["health_timeout"]).seconds *
                SECOND if args["health_timeout"] else None,
                retries=args["health_retries"] if args[
                    "health_retries"] else None,
            )
        else:
            healthcheck = None
        del args["health_cmd"]
        del args["health_interval"]
        del args["health_timeout"]
        del args["health_retries"]
        del args["no_healthcheck"]

        # labels
        labels = dict(args["labels"]) if args["labels"] else None
        del args["labels"]

        # mac_address
        mac_address = args["mac_address"]
        del args["mac_address"]

        # name
        name = args["name"]
        del args["name"]

        # stop_signal
        stop_signal = args["stop_signal"]
        del args["stop_signal"]

        # tty
        tty = args["tty"]
        del args["tty"]

        # volume_driver
        volume_driver = args["volume_driver"]
        del args["volume_driver"]

        # stdin_open
        stdin_open = args["stdin_open"]
        del args["stdin_open"]

        # create_networking_config
        network = args["network"]
        ipv4_address = args["ipv4_address"]
        ipv6_address = args["ipv6_address"]
        aliases = args["aliases"]
        link_local_ips = args["link_local_ips"]
        del args["network"]
        del args["ipv4_address"]
        del args["ipv6_address"]
        del args["aliases"]
        del args["link_local_ips"]
        networking_config = self.client.create_networking_config({
            '{}'.format(network): self.client.create_endpoint_config(
                ipv4_address=ipv4_address,
                ipv6_address=ipv6_address,
                aliases=aliases,
                link_local_ips=link_local_ips
            )
        })

        # host_config
        # wait for PR: https://github.com/docker/docker-py/pull/1363
        del args["cpuset_mems"]
        args["extra_hosts"] = dict(args["extra_hosts"]) if args[
            "extra_hosts"] else None
        if args["links"]:
            for e in args["links"]:
                if len(e) == 1:
                    e.append(e[0])
            args["links"] = dict(args["links"])
        args["restart_policy"] = {
            "Name": args["restart_policy"],
            "MaximumRetryCount": 0
        }
        args["sysctls"] = dict(args["sysctls"]) if args["sysctls"] else None
        args["tmpfs"] = dict(args["tmpfs"]) if args["tmpfs"] else None
        args["ulimits"] = dict(args["ulimits"]) if args["ulimits"] else None
        args["volumes_from"] = dict(args["volumes_from"]) if args[
            "volumes_from"] else None
        host_config = self.client.create_host_config(**args)
        del args["extra_hosts"]
        del args["blkio_weight"]
        del args["cap_add"]
        del args["cap_drop"]
        del args["cgroup_parent"]
        del args["cpu_period"]
        del args["cpu_quota"]
        del args["cpu_shares"]
        del args["cpuset_cpus"]
        # del args["cpuset_mems"]
        del args["devices"]
        del args["dns"]
        del args["dns_opt"]
        del args["dns_search"]
        del args["group_add"]
        del args["ipc_mode"]
        del args["isolation"]
        del args["kernel_memory"]
        del args["links"]
        del args["mem_limit"]
        del args["mem_reservation"]
        del args["memswap_limit"]
        del args["mem_swappiness"]
        del args["oom_kill_disable"]
        del args["oom_score_adj"]
        del args["pid_mode"]
        del args["pids_limit"]
        del args["privileged"]
        del args["publish_all_ports"]
        del args["read_only"]
        del args["restart_policy"]
        del args["security_opt"]
        del args["shm_size"]
        del args["sysctls"]
        del args["tmpfs"]
        del args["ulimits"]
        del args["userns_mode"]

        # store back arguments
        del args["port_bindings"]
        args["ports"] = ports
        del args["binds"]
        args["volumes"] = volumes
        args["networking_config"] = networking_config
        args["command"] = command
        args["host_config"] = host_config
        args["image"] = image
        args["hostname"] = hostname
        args["cidfile"] = cidfile
        args["name"] = name
        args["detach"] = detach
        args["entrypoint"] = entrypoint
        args["environment"] = environment
        args["healthcheck"] = healthcheck
        args["labels"] = labels
        args["mac_address"] = mac_address
        args["name"] = name
        args["stop_signal"] = stop_signal
        args["tty"] = tty
        args["volume_driver"] = volume_driver
        args["stdin_open"] = stdin_open

    def eval_command(self, args):
        """Create host config for containers"""
        try:
            self.preprocess(args)
            # cidfile
            if args["cidfile"]:
                cidfile = args["cidfile"]
            else:
                cidfile = None
            del args["cidfile"]
            self.settings[self.name] = self.client.create_container(**args)
            if cidfile:
                with open(cidfile, 'w') as fd:
                    fd.write(self.settings[self.name]["Id"])
        except APIError as e:
            raise e

    def final(self):
        return self.settings[self.name]



"""This module contains `docker container start` class"""

import dockerpty

from .command import Command


class Start(Command):
    """This class implements `docker container start` command"""

    name = "container start"
    require = ["create"]

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        """Start containers"""
        if (args["detach"] is False) and (args["tty"]):
            dockerpty.start(self.client, self.settings["create"], interactive=args["stdin_open"])
        else:
            self.client.start(self.settings["create"])

    def final(self):
        return self.settings[self.name]



# PYTHON_ARGCOMPLETE_OK
# -*- coding: UTF-8 -*-

from __future__ import absolute_import

from os import path
import argparse
import textwrap
import sys
import argcomplete
from docker.errors import APIError, NotFound
from requests import ConnectionError

from .lib.docker_client import Docker
from .lib.Utils.logger import Logger
from .lib.Utils.switch import switch
from .lib.Utils.shortcut import shortcut


class ArgumentParser(argparse.ArgumentParser):
    """Custom ArgumentParser"""

    def error(self, message):
        self.print_help(sys.stderr)
        raise RuntimeError


class Tsaotun(object):
    """Entrypoint of Tsaotun"""
    color = 0
    debug = 0
    dry = 0
    level = 0
    verbose = 0
    original = 0
    remove = False
    parser = None
    args = None
    buf = None

    class Response(object):
        """Response of tsaotun processing"""
        okay = False
        exception = None
        message = None

        def is_okay(self):
            """Return if response is okay"""
            return self.okay

        def set_exception(self, e):
            """Set exception type"""
            self.exception = e
            self.okay = False

    response = Response()

    def __init__(self, original=False):
        self.__register()
        if original:
            self.set_original()

    def __register(self):
        """Register arguments"""
        # -------------------------START------------------------------
        self.parser = ArgumentParser(
            formatter_class=argparse.RawDescriptionHelpFormatter)
        self.parser.add_argument('--console', '-c',
                                 action="store_true",
                                 help="enter console mode")
        self.parser.add_argument('--color',
                                 action="store_true",
                                 help="set terminal color")
        self.parser.add_argument('--debug',
                                 action="store_true",
                                 help="debug on/off")
        self.parser.add_argument('--dry',
                                 action="store_true",
                                 help="dry-run on/off")
        self.parser.add_argument('--host', '-H',
                                 dest="host",
                                 metavar="list",
                                 help="Daemon socket(s) to connect to (default [])")
        self.parser.add_argument('--verbose', '-v',
                                 action="count", dest="verbosity", default=0,
                                 help="set verbosity level")
        self.parser.add_argument('--original', '-o',
                                 action="store_true", dest="original",
                                 help="original on/off")

        # ------------------------------------------------------------

        sp = self.parser.add_subparsers(
            title="Management Commands", dest="manage_flag")

        # -------------------------VERSION----------------------------

        sp.add_parser('version',
                      formatter_class=argparse.RawDescriptionHelpFormatter,
                      usage="%(prog)s [OPTIONS]",
                      description=textwrap.dedent('''\
        Show the Docker version information
         '''))

        # ---------------------------INFO-----------------------------

        sp.add_parser('info',
                      formatter_class=argparse.RawDescriptionHelpFormatter,
                      usage="%(prog)s",
                      description=textwrap.dedent('''\
        Display system-wide information
         '''))

        # ---------------------------INSPECT--------------------------

        inspect = sp.add_parser(
            'inspect', formatter_class=argparse.RawDescriptionHelpFormatter,
            usage="%(prog)s [OPTIONS] CONTAINER|IMAGE|TASK [CONTAINER|IMAGE|TASK...]",
            description=textwrap.dedent('''\
            Return low-level information on a container, image or task
            '''))
        inspect.add_argument('object',
                             type=str,
                             help="Return low-level information on a container, image or task")
        inspect.add_argument('--format', '-f',
                             type=str,
                             dest="format",
                             help="Pretty-print containers using a Python template")

        # ------------------------CONTAINER---------------------------

        container_p = sp.add_parser('container')
        container = container_p.add_subparsers(
            title="Containers", dest="container_flag", help='type [COMMAND] --help to get additional help')

        # ---------------------CONTAINER-INSPECT----------------------

        container_inspect = container.add_parser(
            'inspect', formatter_class=argparse.RawDescriptionHelpFormatter,
            usage="%(prog)s [OPTIONS] CONTAINER [CONTAINER...]",
            description=textwrap.dedent('''\
            Display detailed information on one or more containers
            '''))
        container_inspect.add_argument('object',
                                       type=str,
                                       help="Display detailed information on one or more containers")
        container_inspect.add_argument('--format', '-f',
                                       type=str,
                                       dest="format",
                                       help="Pretty-print containers using a Python template")

        # -----------------------CONTAINER-LS-------------------------

        container_ls = container.add_parser('ls',
                                            formatter_class=argparse.RawDescriptionHelpFormatter,
                                            usage="%(prog)s [OPTIONS]",
                                            description=textwrap.dedent('''\
        List containers

        Aliases:
          ls, ps, list
         '''))
        container_ls.add_argument('--all', '-a',
                                  action="store_true",
                                  dest="all",
                                  help="Show all containers (default shows just running)")
        container_ls.add_argument('--filter', '-f',
                                  type=dict,
                                  dest="filters",
                                  metavar="filter",
                                  help="Filter output based on conditions provided (default [])")
        container_ls.add_argument('--format',
                                  type=str,
                                  metavar="string",
                                  help="Pretty-print containers using a Python template")
        container_ls.add_argument('--quiet', '-q',
                                  action="store_true",
                                  help="Only display numeric IDs")

        container_ps = container.add_parser('ps',
                                            formatter_class=argparse.RawDescriptionHelpFormatter,
                                            usage="%(prog)s [OPTIONS]",
                                            description=textwrap.dedent('''\
        List containers

        Aliases:
          ls, ps, list
         '''))
        container_ps.add_argument('--all', '-a',
                                  action="store_true",
                                  dest="all",
                                  help="Show all containers (default shows just running)")
        container_ps.add_argument('--filter', '-f',
                                  type=dict,
                                  dest="filters",
                                  metavar="filter",
                                  help="Filter output based on conditions provided (default [])")
        container_ps.add_argument('--format',
                                  type=str,
                                  metavar="string",
                                  help="Pretty-print containers using a Python template")
        container_ps.add_argument('--quiet', '-q',
                                  action="store_true",
                                  help="Only display numeric IDs")

        container_list = container.add_parser('list',
                                              formatter_class=argparse.RawDescriptionHelpFormatter,
                                              usage="%(prog)s [OPTIONS]",
                                              description=textwrap.dedent('''\
        List containers

        Aliases:
          ls, ps, list
         '''))
        container_list.add_argument('--all', '-a',
                                    action="store_true",
                                    dest="all",
                                    help="Show all containers (default shows just running)")
        container_list.add_argument('--filter', '-f',
                                    type=dict,
                                    dest="filters",
                                    metavar="filter",
                                    help="Filter output based on conditions provided (default [])")
        container_list.add_argument('--format',
                                    type=str,
                                    metavar="string",
                                    help="Pretty-print containers using a Python template")
        container_list.add_argument('--quiet', '-q',
                                    action="store_true",
                                    help="Only display numeric IDs")

        # ------------------------CONTAINER-RUN------------------------

        container_run = container.add_parser('run',
                                             add_help=False,
                                             formatter_class=argparse.RawDescriptionHelpFormatter,
                                             usage="%(prog)s [OPTIONS] IMAGE [COMMAND] [ARG...]",
                                             description=textwrap.dedent('''\
        Run a command in a new container
         '''))
        container_run.add_argument('image',
                                   type=str,
                                   metavar="IMAGE",
                                   help="Image to run")
        container_run.add_argument('command',
                                   type=str,
                                   metavar="COMMAND",
                                   nargs="*",
                                   help="Command to run")
        container_run.add_argument('--add-host',
                                   type=lambda kv: kv.split(":", 1),
                                   action="append",
                                   dest="extra_hosts",
                                   metavar="list",
                                   help="Add a custom host-to-IP mapping (host:ip) (default [])")
        container_run.add_argument('--blkio-weight',
                                   type=int,
                                   dest="blkio_weight",
                                   metavar="uint16",
                                   default=0,
                                   help="Block IO (relative weight), between 10 and 1000, or 0 to disable (default 0)")
        container_run.add_argument('--cap-add',
                                   action="append",
                                   type=str,
                                   dest="cap_add",
                                   metavar="list",
                                   help="Add Linux capabilities (default [])")
        container_run.add_argument('--cap-drop',
                                   action="append",
                                   type=str,
                                   dest="cap_drop",
                                   metavar="list",
                                   help="Drop Linux capabilities (default [])")
        container_run.add_argument('--cgroup-parent',
                                   type=str,
                                   dest="cgroup_parent",
                                   metavar="string",
                                   help="Optional parent cgroup for the container")
        container_run.add_argument('--cidfile',
                                   type=str,
                                   dest="cidfile",
                                   metavar="string",
                                   help="Write the container ID to the file")
        container_run.add_argument('--cpu-period',
                                   type=int,
                                   dest="cpu_period",
                                   metavar="int",
                                   help="Limit CPU CFS (Completely Fair Scheduler) period")
        container_run.add_argument('--cpu-quota',
                                   type=int,
                                   dest="cpu_quota",
                                   metavar="int",
                                   help="Limit CPU CFS (Completely Fair Scheduler) quota")
        container_run.add_argument('--cpu-shares', '-c',
                                   type=int,
                                   dest="cpu_shares",
                                   metavar="int",
                                   help="CPU shares (relative weight)")
        container_run.add_argument('--cpuset-cpus',
                                   type=str,
                                   dest="cpuset_cpus",
                                   metavar="string",
                                   help="CPUs in which to allow execution (0-3, 0,1)")
        container_run.add_argument('--cpuset-mems',
                                   type=str,
                                   dest="cpuset_mems",
                                   metavar="string",
                                   help="MEMs in which to allow execution (0-3, 0,1)")
        container_run.add_argument('--detach', '-d',
                                   action="store_true",
                                   dest="detach",
                                   help="Run container in background and print container ID")
        container_run.add_argument('--device',
                                   action="append",
                                   type=str,
                                   dest="devices",
                                   metavar="list",
                                   help="Add a host device to the container (default [])")
        container_run.add_argument('--dns',
                                   action="append",
                                   type=str,
                                   dest="dns",
                                   metavar="list",
                                   help="Set custom DNS servers (default [])")
        container_run.add_argument('--dns-option',
                                   action="append",
                                   type=str,
                                   dest="dns_opt",
                                   metavar="list",
                                   help="Set DNS options (default [])")
        container_run.add_argument('--dns-search',
                                   action="append",
                                   type=str,
                                   dest="dns_search",
                                   metavar="list",
                                   help="Set custom DNS search domains (default [])")
        container_run.add_argument('--entrypoint',
                                   type=str,
                                   dest="entrypoint",
                                   metavar="string",
                                   help="Overwrite the default ENTRYPOINT of the image")
        container_run.add_argument('--env', '-e',
                                   action="append",
                                   type=str,
                                   dest="environment",
                                   metavar="list",
                                   help="Set environment variables (default [])")
        container_run.add_argument('--group-add',
                                   action="append",
                                   type=str,
                                   dest="group_add",
                                   metavar="list",
                                   help="Add additional groups to join (default [])")
        container_run.add_argument('--health-cmd',
                                   type=str,
                                   dest="health_cmd",
                                   metavar="string",
                                   help="Command to run to check health")
        container_run.add_argument('--health-interval',
                                   type=str,
                                   dest="health_interval",
                                   metavar="duration",
                                   help="Time between running the check (ns|us|ms|s|m|h) (default 0s)")
        container_run.add_argument('--health-retries',
                                   type=int,
                                   dest="health_retries",
                                   metavar="int",
                                   help="Consecutive failures needed to report unhealthy")
        container_run.add_argument('--health-timeout',
                                   type=str,
                                   dest="health_timeout",
                                   metavar="duration",
                                   help="Maximum time to allow one check to run (ns|us|ms|s|m|h) (default 0s)")
        container_run.add_argument('--hostname', '-h',
                                   type=str,
                                   dest="hostname",
                                   metavar="string",
                                   help="Container host name")
        container_run.add_argument('--interactive', '-i',
                                   action="store_true",
                                   dest="stdin_open",
                                   help="Keep STDIN open even if not attached")
        container_run.add_argument('--ip',
                                   type=str,
                                   dest="ipv4_address",
                                   metavar="string",
                                   help="Container IPv4 address (e.g. 172.30.100.104)")
        container_run.add_argument('--ip6',
                                   type=str,
                                   dest="ipv6_address",
                                   metavar="string",
                                   help="Container IPv6 address (e.g. 2001:db8::33)")
        container_run.add_argument('--ipc',
                                   type=str,
                                   dest="ipc_mode",
                                   metavar="string",
                                   help="IPC namespace to use")
        container_run.add_argument('--isolation',
                                   type=str,
                                   metavar="string",
                                   help="Container isolation technology")
        container_run.add_argument('--kernel-memory',
                                   type=str,
                                   dest="kernel_memory",
                                   metavar="string",
                                   help="Kernel memory limit")
        container_run.add_argument('--label', '-l',
                                   type=lambda kv: kv.split("=", 1),
                                   action="append",
                                   dest="labels",
                                   metavar="list",
                                   help="Set meta data on a container (default [])")
        container_run.add_argument('--link',
                                   type=lambda kv: kv.split(":", 1),
                                   action="append",
                                   dest="links",
                                   metavar="list",
                                   help="Add link to another container (default [])")
        container_run.add_argument('--link-local-ip',
                                   type=str,
                                   action="append",
                                   dest="link_local_ips",
                                   metavar="list",
                                   help="Container IPv4/IPv6 link-local addresses (default [])")
        container_run.add_argument('--mac-address',
                                   type=str,
                                   dest="mac_address",
                                   metavar="string",
                                   help="Container MAC address (e.g. 92:d0:c6:0a:29:33)")
        container_run.add_argument('--memory', '-m',
                                   type=str,
                                   dest="mem_limit",
                                   metavar="string",
                                   help="Memory limit")
        container_run.add_argument('--memory-reservation',
                                   type=str,
                                   dest="mem_reservation",
                                   metavar="string",
                                   help="Memory soft limit")
        container_run.add_argument('--memory-swap',
                                   type=str,
                                   dest="memswap_limit",
                                   metavar="string",
                                   help="Swap limit equal to memory plus swap: '-1' to enable unlimited swap")
        container_run.add_argument('--memory-swappiness',
                                   type=int,
                                   dest="mem_swappiness",
                                   default=-1,
                                   metavar="int",
                                   help="Tune container memory swappiness (0 to 100) (default -1)")
        container_run.add_argument('--name',
                                   type=str,
                                   metavar="string",
                                   help="Assign a name to the container")
        container_run.add_argument('--network',
                                   type=str,
                                   dest="network",
                                   default="default",
                                   metavar="string",
                                   help="Connect a container to a network (default \"default\")")
        container_run.add_argument('--network-alias',
                                   type=str,
                                   action="append",
                                   dest="aliases",
                                   metavar="list",
                                   help="Add network-scoped alias for the container (default [])")
        container_run.add_argument('--no-healthcheck',
                                   action="store_true",
                                   dest="no_healthcheck",
                                   help="Disable any container-specified HEALTHCHECK")
        container_run.add_argument('--oom-kill-disable',
                                   action="store_true",
                                   dest="oom_kill_disable",
                                   help="Disable OOM Killer")
        container_run.add_argument('--oom-score-adj',
                                   type=int,
                                   dest="oom_score_adj",
                                   metavar="int",
                                   help="Tune host's OOM preferences (-1000 to 1000)")
        container_run.add_argument('--pid',
                                   type=str,
                                   dest="pid_mode",
                                   metavar="string",
                                   help="PID namespace to use")
        container_run.add_argument('--pids-limit',
                                   type=int,
                                   dest="pids_limit",
                                   metavar="int",
                                   help="Tune container pids limit (set -1 for unlimited)")
        container_run.add_argument('--privileged',
                                   action="store_true",
                                   help="Give extended privileges to this container")
        container_run.add_argument('--publish', '-p',
                                   action="append",
                                   type=str,
                                   metavar="list",
                                   dest="port_bindings",
                                   help="Publish a container's port(s) to the host (default [])")
        container_run.add_argument('--publish-all', '-P',
                                   action="store_true",
                                   dest="publish_all_ports",
                                   help="Publish all exposed ports to random ports")
        container_run.add_argument('--read-only',
                                   action="store_true",
                                   dest="read_only",
                                   help="Mount the container's root filesystem as read only")
        container_run.add_argument('--restart',
                                   dest="restart_policy",
                                   default="no",
                                   metavar="string",
                                   help="Restart policy to apply when a container exits (default \"no\")")
        container_run.add_argument('--rm',
                                   action="store_true",
                                   help="Automatically remove the container when it exits")
        container_run.add_argument('--security-opt',
                                   action="append",
                                   dest="security_opt",
                                   metavar="list",
                                   help="Security Options (default [])")
        container_run.add_argument('--shm-size',
                                   dest="shm_size",
                                   default="64MB",
                                   metavar="string",
                                   help="Size of /dev/shm, default value is 64MB")
        container_run.add_argument('--stop-signal',
                                   type=str,
                                   dest="stop_signal",
                                   default="SIGTERM",
                                   metavar="string",
                                   help="Signal to stop a container, SIGTERM by default (default \"SIGTERM\")")
        container_run.add_argument('--sysctl',
                                   type=lambda kv: kv.split("=", 1),
                                   action="append",
                                   dest="sysctls",
                                   metavar="map",
                                   help="Sysctl options (default map[])")
        container_run.add_argument('--tmpfs',
                                   type=lambda kv: kv.split(":", 1),
                                   action="append",
                                   metavar="list",
                                   help="Mount a tmpfs directory (default [])")
        container_run.add_argument('--tty', '-t',
                                   action="store_true",
                                   dest="tty",
                                   help="Allocate a pseudo-TTY")
        container_run.add_argument('--ulimit',
                                   type=lambda kv: kv.split("=", 1),
                                   action="append",
                                   dest="ulimits",
                                   metavar="ulimit",
                                   help="Ulimit options (default [])")
        container_run.add_argument('--userns',
                                   type=str,
                                   dest="userns_mode",
                                   metavar="string",
                                   help="User namespace to use")
        container_run.add_argument('--volume', '-v',
                                   action="append",
                                   type=str,
                                   dest="binds",
                                   help=textwrap.dedent('''\
        Bind mount a volume (default []). The format
        is [host-src:]container-dest[:<options>].
        The comma-delimited options are [rw|ro],
        [z|Z], [[r]shared|[r]slave|[r]private], and
        [nocopy]. The 'host-src' is an absolute path
        or a name value.
        '''))
        container_run.add_argument('--volumes-driver',
                                   type=str,
                                   dest="volume_driver",
                                   metavar="string",
                                   help="Optional volume driver for the container")
        container_run.add_argument('--volumes-from',
                                   type=lambda kv: kv.split(":", 1),
                                   action="append",
                                   dest="volumes_from",
                                   metavar="list",
                                   help="Mount volumes from the specified container(s) (default [])")
        container_run.add_argument('--help', action='help',
                                   help='show this help message and exit')

        # --------------------CONTAINER-INSPECT-----------------------

        container_inspect = container.add_parser(
            'inspect', formatter_class=argparse.RawDescriptionHelpFormatter,
            usage="%(prog)s [OPTIONS] CONTAINER|IMAGE|TASK [CONTAINER|IMAGE|TASK...]",
            description=textwrap.dedent('''\
            Return low-level information on a container, image or task
            '''))
        container_inspect.add_argument('object',
                                       type=str,
                                       help="Return low-level information on a container, image or task")
        container_inspect.add_argument('--format', '-f',
                                       type=str,
                                       dest="format",
                                       help="Pretty-print containers using a Python template")

        # ----------------------CONTAINER-LOGS------------------------

        container_logs = container.add_parser('logs',
                                              formatter_class=argparse.RawDescriptionHelpFormatter,
                                              usage="%(prog)s [OPTIONS] CONTAINER",
                                              description=textwrap.dedent('''\
        Fetch the logs of a container
         '''))
        container_logs.add_argument('container',
                                    type=str,
                                    metavar="CONTAINER",
                                    help="Fetch the logs of a container")

        # ---------------------CONTAINER-STATS------------------------

        container_stats = container.add_parser('stats',
                                               formatter_class=argparse.RawDescriptionHelpFormatter,
                                               usage="%(prog)s [OPTIONS] [CONTAINER...]",
                                               description=textwrap.dedent('''\
        Display a live stream of container(s) resource usage statistics
         '''))
        container_stats.add_argument('containers',
                                     type=str,
                                     metavar="CONTAINER",
                                     nargs="+",
                                     help="Display a live stream of container(s) resource usage statistics")
        """
        container_stats.add_argument('--all', '-a',
                           action="store_true",
                           default=False,
                           help="Show all containers (default shows just running)")
        """
        container_stats.add_argument('--no-stream',
                                     action="store_true",
                                     dest="stream",
                                     default=True,
                                     help="Disable streaming stats and only pull the first result")

        # --------------------CONTAINER-RENAME------------------------

        container_rename = container.add_parser('rename',
                                                formatter_class=argparse.RawDescriptionHelpFormatter,
                                                usage="%(prog)s CONTAINER NEW_NAME",
                                                description=textwrap.dedent('''\
        Rename a container
         '''))
        container_rename.add_argument('container',
                                      type=str,
                                      metavar="CONTAINER",
                                      help="Container to be renamed")
        container_rename.add_argument('name',
                                      type=str,
                                      metavar="NAME",
                                      help="New container name")

        # --------------------CONTAINER-RESTART-------------------------

        container_restart = container.add_parser('restart',
                                                 formatter_class=argparse.RawDescriptionHelpFormatter,
                                                 usage="%(prog)s [OPTIONS] CONTAINER [CONTAINER...]",
                                                 description=textwrap.dedent('''\
        Restart one or more containers
         '''))
        container_restart.add_argument('containers',
                                       type=str,
                                       metavar="CONTAINER",
                                       nargs="+",
                                       help="Containers to be restarted")
        container_restart.add_argument('--time', '-t',
                                       type=int,
                                       metavar='int',
                                       dest="timeout",
                                       default=10,
                                       help="Seconds to wait for stop before killing the container (default 10)")

        # ---------------------CONTAINER-EXEC-------------------------

        container_exec = container.add_parser('exec',
                                              formatter_class=argparse.RawDescriptionHelpFormatter,
                                              usage="%(prog)s [OPTIONS] CONTAINER COMMAND [ARG...]",
                                              description=textwrap.dedent('''\
        Run a command in a running container
         '''))
        container_exec.add_argument('container',
                                    type=str,
                                    metavar="CONTAINER",
                                    help="Target container where exec instance will be created")
        container_exec.add_argument('cmd',
                                    type=str,
                                    metavar="COMMAND",
                                    nargs="+",
                                    help="Command to be executed")
        container_exec.add_argument('--detach', '-d',
                                    action="store_true",
                                    dest="detach",
                                    help="Detached mode: run command in the background")
        container_exec.add_argument('--interactive', '-i',
                                    action="store_true",
                                    dest="interactive",
                                    default=True,
                                    help="Keep STDIN open even if not attached")
        container_exec.add_argument('--tty', '-t',
                                    action="store_true",
                                    dest="tty",
                                    help="Allocate a pseudo-TTY")

        # -----------------------CONTAINER-CP-------------------------

        container_cp = container.add_parser('cp',
                                            formatter_class=argparse.RawDescriptionHelpFormatter,
                                            usage="%(prog)s [OPTIONS] CONTAINER:SRC_PATH DEST_PATH|-\ndocker cp [OPTIONS] SRC_PATH|- CONTAINER:DEST_PATH",
                                            description=textwrap.dedent('''\
        Copy files/folders between a container and the local filesystem

        Use '-' as the source to read a tar archive from stdin
        and extract it to a directory destination in a container.
        Use '-' as the destination to stream a tar archive of a
        container source to stdout.
         '''))
        container_cp.add_argument('src',
                                  type=str,
                                  metavar="CONTAINER:SRC_PATH")
        container_cp.add_argument('dest',
                                  type=str,
                                  metavar="DEST_PATH")

        # ----------------------CONTAINER-RM--------------------------

        container_rm = container.add_parser('rm',
                                            formatter_class=argparse.RawDescriptionHelpFormatter,
                                            usage="%(prog)s [OPTIONS] CONTAINER [CONTAINER...]",
                                            description=textwrap.dedent('''\
        Remove one or more containers
         '''))
        container_rm.add_argument('containers',
                                  type=str,
                                  metavar="CONTAINER",
                                  nargs="+",
                                  help="Containers to be removed")
        container_rm.add_argument('--force', '-f',
                                  action="store_true",
                                  dest="force",
                                  help="Force the removal of a running container (uses SIGKILL)")

        # -----------------------CONTAINER-TOP------------------------

        container_top = container.add_parser('top',
                                             formatter_class=argparse.RawDescriptionHelpFormatter,
                                             usage="%(prog)s CONTAINER [ps OPTIONS]",
                                             description=textwrap.dedent('''\
        Display the running processes of a container
         '''))
        container_top.add_argument('container',
                                   type=str,
                                   metavar="CONTAINER",
                                   help="Display the running processes of a container")
        container_top.add_argument('ps_args',
                                   type=str,
                                   nargs="?",
                                   help="Ps options")

        # --------------------------IMAGE-----------------------------

        image_p = sp.add_parser('image')
        image = image_p.add_subparsers(
            title="Images", dest="image_flag", help='type [COMMAND] --help to get additional help')

        # -----------------------IMAGE-INSPECT------------------------

        image_inspect = image.add_parser(
            'inspect', formatter_class=argparse.RawDescriptionHelpFormatter,
            usage="%(prog)s [OPTIONS] IMAGE [IMAGE...]",
            description=textwrap.dedent('''\
            Display detailed information on one or more images
            '''))
        image_inspect.add_argument('object',
                                   type=str,
                                   help="Display detailed information on one or more images")
        image_inspect.add_argument('--format', '-f',
                                   type=str,
                                   dest="format",
                                   help="Pretty-print containers using a Python template")

        # -------------------------IMAGE-LS---------------------------

        image_ls = image.add_parser('ls',
                                    formatter_class=argparse.RawDescriptionHelpFormatter,
                                    usage="%(prog)s [OPTIONS] [REPOSITORY[:TAG]]",
                                    description=textwrap.dedent('''\
        List images

        Aliases:
          ls, images, list
         '''))
        image_ls.add_argument('--all', '-a',
                              action="store_true",
                              dest="all",
                              help="Show all images (default hides intermediate images)")
        image_ls.add_argument('--digests',
                              action="store_true",
                              dest="digests",
                              help="Show digests")
        image_ls.add_argument('--filter', '-f',
                              type=lambda kv: kv.split("=", 1),
                              action="append",
                              dest="filters",
                              help="Filter output based on conditions provided")
        image_ls.add_argument('--format',
                              type=str,
                              help="Pretty-print containers using a Python template")
        image_ls.add_argument('--quiet', '-q',
                              action="store_true",
                              help="Only show numeric IDs")

        image_images = image.add_parser('images',
                                        formatter_class=argparse.RawDescriptionHelpFormatter,
                                        usage="%(prog)s [OPTIONS] [REPOSITORY[:TAG]]",
                                        description=textwrap.dedent('''\
        List images

        Aliases:
          ls, images, list
         '''))
        image_images.add_argument('--all', '-a',
                                  action="store_true",
                                  dest="all",
                                  help="Show all images (default hides intermediate images)")
        image_images.add_argument('--digests',
                                  action="store_true",
                                  dest="digests",
                                  help="Show digests")
        image_images.add_argument('--filter', '-f',
                                  type=lambda kv: kv.split("=", 1),
                                  action="append",
                                  dest="filters",
                                  help="Filter output based on conditions provided")
        image_images.add_argument('--format',
                                  type=str,
                                  help="Pretty-print containers using a Python template")
        image_images.add_argument('--quiet', '-q',
                                  action="store_true",
                                  help="Only show numeric IDs")

        image_list = image.add_parser('list',
                                      formatter_class=argparse.RawDescriptionHelpFormatter,
                                      usage="%(prog)s [OPTIONS] [REPOSITORY[:TAG]]",
                                      description=textwrap.dedent('''\
        List images

        Aliases:
          ls, images, list
         '''))
        image_list.add_argument('--all', '-a',
                                action="store_true",
                                dest="all",
                                help="Show all images (default hides intermediate images)")
        image_list.add_argument('--digests',
                                action="store_true",
                                dest="digests",
                                help="Show digests")
        image_list.add_argument('--filter', '-f',
                                type=lambda kv: kv.split("=", 1),
                                action="append",
                                dest="filters",
                                help="Filter output based on conditions provided")
        image_list.add_argument('--format',
                                type=str,
                                help="Pretty-print containers using a Python template")
        image_list.add_argument('--quiet', '-q',
                                action="store_true",
                                help="Only show numeric IDs")

        # ------------------------IMAGE-PULL--------------------------

        image_pull = image.add_parser('pull',
                                      formatter_class=argparse.RawDescriptionHelpFormatter,
                                      usage="%(prog)s [OPTIONS] NAME[:TAG|@DIGEST]",
                                      description=textwrap.dedent('''\
        Pull an image or a repository from a registry
         '''))
        image_pull.add_argument('image',
                                type=str,
                                metavar="IMAGE",
                                nargs=1,
                                help="Image to pull")

        # ------------------------IMAGE-BUILD-------------------------

        image_pull = image.add_parser('build',
                                      formatter_class=argparse.RawDescriptionHelpFormatter,
                                      usage="%(prog)s [OPTIONS] PATH | URL | -",
                                      description=textwrap.dedent('''\
        Build an image from a Dockerfile
         '''))
        image_pull.add_argument('path',
                                type=str,
                                metavar="PATH",
                                help="The path containing Dockerfile")
        image_pull.add_argument('--build-arg',
                                type=lambda kv: kv.split("=", 1),
                                action="append",
                                dest="buildargs",
                                metavar="list",
                                help="Set build-time variables (default [])")
        image_pull.add_argument('--tag', '-t',
                                type=str,
                                dest="tag",
                                help="Name and optionally a tag in the 'name:tag' format")
        image_pull.add_argument('--rm',
                                action="store_true",
                                default=True,
                                help="Remove intermediate containers after a successful build (default true)")

        # ------------------------IMAGE-SAVE--------------------------

        image_save = image.add_parser('save',
                                      formatter_class=argparse.RawDescriptionHelpFormatter,
                                      usage="%(prog)s [OPTIONS] IMAGE [IMAGE...]",
                                      description=textwrap.dedent('''\
        Save one or more images to a tar archive (streamed to STDOUT by default)
         '''))
        image_save.add_argument('image',
                                type=str,
                                metavar="IMAGE",
                                help="Image to be saved")

        image_save.add_argument('--output', '-o',
                                type=str,
                                dest="output",
                                metavar="string",
                                help="Write to a file, instead of STDOUT")

        # -------------------------IMAGE-RM---------------------------

        image_rm = image.add_parser('rm',
                                    formatter_class=argparse.RawDescriptionHelpFormatter,
                                    usage="%(prog)s [OPTIONS] IMAGE [IMAGE...]",
                                    description=textwrap.dedent('''\
        Remove one or more images

        Aliases:
          rm, rmi, remove
         '''))
        image_rm.add_argument('images',
                              type=str,
                              metavar="IMAGE",
                              nargs="+",
                              help="Images to be removed")
        image_rm.add_argument('--force', '-f',
                              action="store_true",
                              dest="force",
                              help="Force removal of the image")

        image_rmi = image.add_parser('rmi',
                                     formatter_class=argparse.RawDescriptionHelpFormatter,
                                     usage="%(prog)s [OPTIONS] IMAGE [IMAGE...]",
                                     description=textwrap.dedent('''\
        Remove one or more images

        Aliases:
          rm, rmi, remove
         '''))
        image_rmi.add_argument('images',
                               type=str,
                               metavar="IMAGE",
                               nargs="+",
                               help="Images to be removed")
        image_rmi.add_argument('--force', '-f',
                               action="store_true",
                               dest="force",
                               help="Force removal of the image")

        image_remove = image.add_parser('remove',
                                        formatter_class=argparse.RawDescriptionHelpFormatter,
                                        usage="%(prog)s [OPTIONS] IMAGE [IMAGE...]",
                                        description=textwrap.dedent('''\
        Remove one or more images

        Aliases:
          rm, rmi, remove
         '''))
        image_remove.add_argument('images',
                                  type=str,
                                  metavar="IMAGE",
                                  nargs="+",
                                  help="Images to be removed")
        image_remove.add_argument('--force', '-f',
                                  action="store_true",
                                  dest="force",
                                  help="Force removal of the image")

        # -----------------------IMAGE-HISTORY------------------------

        image_history = image.add_parser('history',
                                         formatter_class=argparse.RawDescriptionHelpFormatter,
                                         usage="%(prog)s [OPTIONS] IMAGE",
                                         description=textwrap.dedent('''\
        Show the history of an image
         '''))
        image_history.add_argument('image',
                                   type=str,
                                   metavar="IMAGE",
                                   help="Image to show history")

        # ------------------------NETWORK-----------------------------

        network_p = sp.add_parser('network')
        network = network_p.add_subparsers(
            title="Networks", dest="network_flag", help='type [COMMAND] --help to get additional help')

        # -----------------------NETWORK-LS---------------------------

        network_ls = network.add_parser('ls',
                                        formatter_class=argparse.RawDescriptionHelpFormatter,
                                        usage="%(prog)s [OPTIONS]",
                                        description=textwrap.dedent('''\
        List networks
         '''))
        network_ls.add_argument('--filter', '-f',
                                type=lambda kv: kv.split("=", 1),
                                action="append",
                                dest="filters",
                                metavar="filter",
                                help="Provide filter values (e.g. 'driver=bridge')")
        network_ls.add_argument('--format',
                                type=str,
                                help="Pretty-print networks using a Python template")

        # ---------------------NETWORK-CREATE--------------------------

        network_create = network.add_parser('create',
                                            formatter_class=argparse.RawDescriptionHelpFormatter,
                                            usage="%(prog)s [OPTIONS] NETWORK",
                                            description=textwrap.dedent('''\
        Create a network
         '''))
        network_create.add_argument('name',
                                    type=str,
                                    metavar="NETWORK",
                                    help="Network name")
        network_create.add_argument('--driver', '-d',
                                    type=str,
                                    dest="driver",
                                    default="bridge",
                                    metavar="string",
                                    help="Driver to manage the Network (default \"bridge\")")
        network_create.add_argument('--gateway',
                                    type=str,
                                    dest="gateway",
                                    help="Custom IP address for the pool's gateway.")
        network_create.add_argument('--internal',
                                    action="store_true",
                                    default=False,
                                    help="Restrict external access to the network")
        network_create.add_argument('--ip-range',
                                    type=str,
                                    dest="iprange",
                                    help="Allocate container ip from a sub-range")
        network_create.add_argument('--ipam-driver',
                                    type=str,
                                    dest="ipamdriver",
                                    default="default",
                                    help="IP Address Management Driver (default \"default\")")
        network_create.add_argument('--ipam-opt',
                                    type=lambda kv: kv.split("=", 1),
                                    action="append",
                                    dest="ipamopt",
                                    help="Set IPAM driver specific options (default None)")
        network_create.add_argument('--ipv6',
                                    action="store_true",
                                    dest="enable_ipv6",
                                    default=False,
                                    help="Enable IPv6 networking")
        network_create.add_argument('--label',
                                    type=lambda kv: kv.split("=", 1),
                                    action="append",
                                    dest="labels",
                                    help="Set metadata on a network (default [])")
        network_create.add_argument('--opt', '-o',
                                    type=lambda kv: kv.split("=", 1),
                                    action="append",
                                    dest="options",
                                    help="Set driver specific options")
        network_create.add_argument('--subnet',
                                    type=str,
                                    dest="subnet",
                                    help="Custom subnet for this IPAM pool using the CIDR notation. Defaults to None.")

        # -----------------------NETWORK-RM----------------------------

        network_rm = network.add_parser(
            'rm', formatter_class=argparse.RawDescriptionHelpFormatter,
            usage="%(prog)s NETWORK [NETWORK...]",
            description=textwrap.dedent('''\
        Remove one or more networks

        Aliases:
          rm, remove
        '''))
        network_rm.add_argument('networks',
                                type=str,
                                metavar="NETWORK",
                                nargs="+",
                                help="Networks to be removed")

        network_remove = network.add_parser(
            'remove', formatter_class=argparse.RawDescriptionHelpFormatter,
            usage="%(prog)s NETWORK [NETWORK...]",
            description=textwrap.dedent('''\
            Remove one or more networks

            Aliases:
              rm, remove
            '''))

        network_remove.add_argument('networks',
                                    type=str,
                                    metavar="NETWORK",
                                    nargs="+",
                                    help="Remove one or more networks")

        # -----------------------NETWORK-INSPECT-----------------------

        network_inspect = network.add_parser(
            'inspect', formatter_class=argparse.RawDescriptionHelpFormatter,
            usage="%(prog)s [OPTIONS] NETWORK [NETWORK...]",
            description=textwrap.dedent('''\
            Display detailed information on one or more networks
            '''))
        network_inspect.add_argument('networks',
                                     type=str,
                                     metavar="NETWORK",
                                     nargs="+",
                                     help="Display detailed information on one or more networks")
        network_inspect.add_argument('--format', '-f',
                                     type=str,
                                     dest="format",
                                     help="Pretty-print containers using a Python template")

        # ------------------------NETWORK-CONNECT-----------------------

        network_connect = network.add_parser(
            'connect', formatter_class=argparse.RawDescriptionHelpFormatter,
            usage="%(prog)s [OPTIONS] NETWORK CONTAINER",
            description=textwrap.dedent('''\
            Connect a container to a network
            '''))
        network_connect.add_argument('net_id',
                                     type=str,
                                     metavar="NETWORK",
                                     help="Network to be connected")
        network_connect.add_argument('container',
                                     type=str,
                                     metavar="CONTAINER",
                                     help="Container to be assigned")
        network_connect.add_argument('--alias',
                                     action="append",
                                     type=str,
                                     dest="aliases",
                                     metavar="stringSlice",
                                     help="Add network-scoped alias for the container")
        network_connect.add_argument('--ip',
                                     type=str,
                                     dest="ipv4_address",
                                     metavar="string",
                                     help="IP Address")
        network_connect.add_argument('--ip6',
                                     type=str,
                                     dest="ipv6_address",
                                     metavar="string",
                                     help="IPv6 Address")
        network_connect.add_argument('--link',
                                     action="append",
                                     type=str,
                                     dest="links",
                                     metavar="list",
                                     help="Add link to another container (default [])")
        network_connect.add_argument('--link-local-ip',
                                     action="append",
                                     type=str,
                                     dest="link_local_ips",
                                     metavar="stringSlice",
                                     help="Add a link-local address for the container")

        # ------------------------NETWORK-DISCONNECT--------------------

        network_disconnect = network.add_parser(
            'disconnect', formatter_class=argparse.RawDescriptionHelpFormatter,
            usage="%(prog)s [OPTIONS] NETWORK CONTAINER",
            description=textwrap.dedent('''\
            Disconnect a container from a network
            '''))
        network_disconnect.add_argument('net_id',
                                        type=str,
                                        metavar="NETWORK",
                                        help="Network to be disconnected")
        network_disconnect.add_argument('container',
                                        type=str,
                                        metavar="CONTAINER",
                                        help="Container to be assigned")
        network_disconnect.add_argument('--force', '-f',
                                        action="store_true",
                                        dest="force",
                                        help="Force the container to disconnect from a network")

        # ---------------------------NETWORK-PRUNE----------------------

        """
        network_prune = network.add_parser(
            'prune', formatter_class=argparse.RawDescriptionHelpFormatter,
            usage="%(prog)s [OPTIONS]",
            description=textwrap.dedent('''\
            Remove all unused networks
            '''))
        network_prune.add_argument('--force', '-f',
                                   action="store_true",
                                   dest="force",
                                   help="Do not prompt for confirmation")
        """

        # -------------------------VOLUME-----------------------------

        volume_p = sp.add_parser('volume')
        volume = volume_p.add_subparsers(
            title="Volumes", dest="volume_flag", help='type [COMMAND] --help to get additional help')

        # ------------------------VOLUME-LS---------------------------

        volume_ls = volume.add_parser('ls',
                                      formatter_class=argparse.RawDescriptionHelpFormatter,
                                      usage="%(prog)s [OPTIONS]",
                                      description=textwrap.dedent('''\
        List volumes
         '''))
        volume_ls.add_argument('--filter', '-f',
                               type=lambda kv: kv.split("=", 1),
                               action="append",
                               dest="filters",
                               metavar="filter",
                               help="Provide filter values (e.g. 'dangling=true')")
        volume_ls.add_argument('--format',
                               type=str,
                               help="Pretty-print networks using a Python template")

        # ----------------------VOLUME-CREATE--------------------------

        volume_create = volume.add_parser('create',
                                          formatter_class=argparse.RawDescriptionHelpFormatter,
                                          usage="%(prog)s [OPTIONS] [VOLUME]",
                                          description=textwrap.dedent('''\
        Create a volume
         '''))
        volume_create.add_argument('name',
                                   type=str,
                                   metavar="VOLUME",
                                   help="Volume name")
        volume_create.add_argument('--driver', '-d',
                                   type=str,
                                   dest="driver",
                                   default="local",
                                   metavar="string",
                                   help="Specify volume driver name (default \"local\")")
        volume_create.add_argument('--opt', '-o',
                                   type=lambda kv: kv.split("=", 1),
                                   action="append",
                                   dest="driver_opts",
                                   metavar="map",
                                   help="Set driver specific options (default map[])")
        volume_create.add_argument('--label',
                                   type=lambda kv: kv.split("=", 1),
                                   action="append",
                                   dest="labels",
                                   metavar="list",
                                   help="Set metadata for a volume (default [])")

        # ------------------------VOLUME-RM----------------------------

        volume_rm = volume.add_parser('rm',
                                      formatter_class=argparse.RawDescriptionHelpFormatter,
                                      usage="%(prog)s [OPTIONS] VOLUME [VOLUME...]",
                                      description=textwrap.dedent('''\
        Remove one or more volumes

        Aliases:
          rm, remove

        Examples:

        $ docker volume rm hello
        hello
         '''))
        volume_rm.add_argument('volumes',
                               type=str,
                               metavar="VOLUME",
                               nargs="+",
                               help="Volumes to be removed")

        volume_remove = volume.add_parser('remove',
                                          formatter_class=argparse.RawDescriptionHelpFormatter,
                                          usage="%(prog)s [OPTIONS] VOLUME [VOLUME...]",
                                          description=textwrap.dedent('''\
        Remove one or more volumes

        Aliases:
          rm, remove

        Examples:

        $ docker volume rm hello
        hello
         '''))
        volume_remove.add_argument('volumes',
                                   type=str,
                                   metavar="VOLUME",
                                   nargs="+",
                                   help="Volumes to be removed")

        # ------------------------VOLUME-INSPECT-----------------------

        volume_inspect = volume.add_parser(
            'inspect', formatter_class=argparse.RawDescriptionHelpFormatter,
            usage="%(prog)s [OPTIONS] VOLUME [VOLUME...]",
            description=textwrap.dedent('''\
            Display detailed information on one or more volumes
            '''))
        volume_inspect.add_argument('volumes',
                                    type=str,
                                    metavar="VOLUME",
                                    nargs="+",
                                    help="Display detailed information on one or more volumes")
        volume_inspect.add_argument('--format', '-f',
                                    type=str,
                                    dest="format",
                                    help="Pretty-print containers using a Python template")

        # ---------------------------END------------------------------

        argcomplete.autocomplete(self.parser)

    def __parse(self, argv):
        """Parse argv into list of args"""
        argv = argv.split() if argv else sys.argv[1:]
        self.args = self.parser.parse_args(shortcut(argv))
        self.args = vars(self.args)

        self.set_verbose(self.args["verbosity"])
        self.set_docker(self.args["host"])
        del self.args["host"]
        if self.args["color"]:
            self.set_color()
        if self.args["debug"]:
            self.set_debug()
        if self.args["dry"]:
            self.set_dry()
        if self.args["original"]:
            self.set_original()
        if self.original == 0:
            self.__plugin()

    def __plugin(self):
        """Setup plugins"""
        import pkgutil
        import imp
        plugins_dir = "{}/Tsaotun/plugins/".format(path.expanduser("~"))
        if path.exists(plugins_dir):
            from .lib.Utils.deepgetattr import deepgetattr
            module_names = [n for _, n, _ in pkgutil.iter_modules(
                ["{}".format(plugins_dir)])]
            tmp = {}
            for module_name in module_names:
                try:
                    f, filename, description = imp.find_module(
                        module_name, [plugins_dir])
                    mod = imp.load_module(
                        module_name, f, filename, description)
                    for k, v in mod.__override__.iteritems():
                        tmp["{}|{}".format(k, v)] = deepgetattr(
                            mod, "{}.{}".format(k, v))
                except ImportError as e:
                    Logger.logError(e)
            self.push(**tmp)

    def set_docker(self, host):
        """Set docker host"""
        self.docker = Docker(host)

    def set_color(self):
        """Set terminal color"""
        self.color = 1

    def set_debug(self):
        """DEBUG on/off"""
        self.debug = 1

    def set_dry(self):
        """dry-run on/off"""
        self.dry = 1

    def set_original(self):
        """Set original on/off"""
        self.original = 1

    def set_verbose(self, level):
        """Set verbosity level"""
        self.verbose = min(level, 3)

    def push(self, **kwargs):
        """Push custom classes to docker"""
        self.docker.push(**kwargs)

    def send(self, argv, suppress=False):
        """Evaluate commands"""
        try:
            self.__parse(argv)
            if self.debug:
                import json
                print json.dumps(self.args, indent=4)
            manage_flag = self.args["manage_flag"]
            command_flag = self.args.get("container_flag")
            del self.args["console"]
            del self.args["color"]
            del self.args["debug"]
            del self.args["dry"]
            del self.args["verbosity"]
            del self.args["original"]
            if (manage_flag == "container") and ("rm" in self.args):
                if (command_flag is not None) and (command_flag == "run"):
                    self.remove = self.args["rm"]
                del self.args["rm"]
            else:
                self.remove = False

            self.docker.load(self.args, dry=self.dry)
            if not self.dry:
                if self.remove:
                    self.docker.client.remove_container(
                        self.docker.buffer()["container create"]["Id"], force=True)
                if (manage_flag == "container") and (command_flag is not None) and (command_flag == "run"):
                    self.docker.set_buffer(
                        self.docker.buffer()["container run"])
            if suppress is not True:
                self.buf = self.docker.buffer()
        except (KeyboardInterrupt, SystemExit) as e:
            self.response.set_exception(type(e))
        except ConnectionError as e:
            self.response.set_exception(type(e))
            self.response.message = e.message[0]
        except (AttributeError, ValueError) as e:
            self.response.set_exception(type(e))
            self.response.message = str(e)
        except (APIError, NotFound) as e:
            self.response.set_exception(type(e))
            try:
                self.response.message = e.response.json()["message"]
            except ValueError:
                self.response.message = str(e.response.text[:-1])
        except RuntimeError as e:
            self.response.set_exception(type(e))

    def recv(self):
        """Return the buffer of tsaotun"""
        return self.buf

    def final(self):
        """Finalize the result"""
        if self.response.exception:
            for case in switch(self.response.exception):
                if case(ConnectionError, AttributeError, ValueError):
                    if self.color:
                        Logger.logError(
                            "Error response from tsaotun: {}\n", self.response.message)
                    else:
                        Logger.log(
                            "Error response from tsaotun: {}\n", self.response.message)
                    break
                if case(APIError, NotFound):
                    if self.color:
                        Logger.logError(
                            "Error response from daemon: {}\n", self.response.message)
                    else:
                        Logger.log(
                            "Error response from daemon: {}\n", self.response.message)
                    break
                if case(SystemExit, KeyboardInterrupt, RuntimeError):
                    pass
        else:
            if self.color:
                Logger.logSuccess("{}", self.recv())
            else:
                Logger.log("{}", self.recv())


def cli(argv=None, **intruders):
    """Entry point of tsaotun"""
    try:
        tsaotun = Tsaotun(**intruders)
        tsaotun.send(argv)
        tsaotun.final()
    except RuntimeError:
        pass



"""This module contains `docker container ls` class"""

import pystache
import arrow
from pytabwriter import TabWriter

from .command import Command


class Ls(Command):
    """This class implements `docker container ls` command"""

    name = "container ls"
    require = []

    defaultTemplate = '{{Id}}\t{{Image}}\t"{{Command}}"\t{{Created}}\t{{Status}}\t{{Ports}}\t{{Names}}'

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def process_ports(self, ports):
        port_list = {"tcp": [], "udp": []}
        for port in ports:
            port_list[port["Type"]].append(str(port["PrivatePort"]))
        for port_type, port_value in port_list.iteritems():
            if len(port_value) == 0:
                del port_value

        return ' '.join(["{}/{}".format(','.join(port_list[port_type]), port_type) for port_type in port_list.keys()])

    def eval_command(self, args):
        tw = TabWriter()
        if args["format"] is None:
            tw.padding = [8, 8, 3, 7, 8, 11]
            fm = self.defaultTemplate
            tw.writeln(
                "CONTAINER ID\tIMAGE\tCOMMAND\tCREATED\tSTATUS\tPORTS\tNAMES")
        else:
            fm = args["format"]
        del args["format"]

        nodes = self.client.containers(**args)
        for node in nodes:
            node["Id"] = node["Id"][:12]
            node["Command"] = (node["Command"][
                :17] + "...") if len(node["Command"]) >= 20 else node["Command"]
            node["Created"] = arrow.get(node["Created"]).humanize()
            node["Ports"] = self.process_ports(node['Ports'])
            node["Names"] = ', '.join([e[1:] for e in node["Names"]])
            tw.writeln(pystache.render(fm, node))
        self.settings[self.name] = str(tw)

    def final(self):
        return self.settings[self.name]



"""This module contains `docker volume create` class"""


from docker.errors import APIError

from .command import Command


class Create(Command):
    """This class implements `docker volume create` command"""

    name = "volume create"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        """Create host config for containers"""
        try:
            args["driver_opts"] = dict(args["driver_opts"]) if args["driver_opts"] else None
            args["labels"] = dict(args["labels"]) if args["labels"] else None
            self.settings[self.name] = self.client.create_volume(**args)["Name"]
        except APIError as e:
            raise e

    def final(self):
        return self.settings[self.name]



"""This module contains `docker container rename` class"""

from .command import Command


class Rename(Command):
    """This class implements `docker container rename` command"""

    name = "container rename"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        self.client.rename(**args)
        self.settings[self.name] = "\r"

    def final(self):
        return self.settings[self.name]



"""This module provides shortcut functionality (mapping command to another command)"""

mapping = {
    "ps": "container ps",
    "images": "image images",
    "pull": "image pull",
    "build": "image build",
    "save": "image save",
    "rmi": "image rm",
    "run": "container run",
    "exec": "container exec",
    "stop": "container stop",
    "restart": "container restart",
    "rename": "container rename",
    "rm": "container rm",
    "cp": "container cp",
    "logs": "container logs",
    "top": "container top"
}


def shortcut(argv):
    """Return mapping command if there exists one"""
    c = 0
    for i, val in enumerate(argv):
        if val.startswith("-"):
            if val == '-H' or val == '--host':
                c = c - 1
            continue
        else:
            c = c + 1
        if (c == 1) and mapping.get(val):
            argv[i] = mapping.get(val)
            break
    return ' '.join(argv).split()



"""This module contains `docker container cp` class"""

import os
import sys
import tempfile
import tarfile
from enum import Enum
from docker.errors import APIError
from ...Utils.switch import switch

from .command import Command


Direction = Enum(
    'Direction', 'fromContainer toContainer acrossContainers', start=0)


class Cp(Command):
    """This class implements `docker container cp` command"""

    name = "container cp"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def split_cp_arg(self, arg):
        """Split provided cp arguments"""
        if os.path.isabs(arg):
            return "", arg

        parts = arg.split(':', 2)

        if (len(parts) == 1) or (parts[0][0] == '.'):
            return "", arg
        else:
            return parts[0], parts[1]

    def copy_from_container(self, args, src_container, src_path, dst_path):
        """Preprocess arguments and tarfile provided for `docker cp`"""
        args['container'] = src_container
        args['path'] = src_path
        tar, stat = self.client.get_archive(**args)
        if dst_path == '-':
            self.settings[self.name] = tar.read()
        else:
            fd = tempfile.NamedTemporaryFile(delete=False)
            fd.write(tar.read())
            fd.close()
            with tarfile.open(fd.name, 'r') as t:
                t.extractall(path=dst_path)
            os.unlink(fd.name)
            self.settings[self.name] = stat

    def copy_to_container(self, args, src_path, dst_container, dst_path):
        """Preprocess arguments and tarfile provided for `docker cp`"""
        args['container'] = dst_container
        args['path'] = dst_path
        if src_path == '-':
            args['data'] = sys.stdin.read()
            if self.client.put_archive(**args):
                self.settings[self.name] = "Done!"
            else:
                self.settings[self.name] = "Failed!"
        else:
            fd = tempfile.NamedTemporaryFile(delete=False)
            with tarfile.open(fd.name, 'w') as t:
                t.add(src_path, os.path.basename(src_path))
            fd.close()
            with open(fd.name, 'rb') as f:
                args['data'] = f.read()
                if self.client.put_archive(**args):
                    self.settings[self.name] = "Done!"
                else:
                    self.settings[self.name] = "Failed!"
            os.unlink(fd.name)

    def eval_command(self, args):
        try:
            direction = 0
            src_container, src_path = self.split_cp_arg(args['src'])
            dst_container, dst_path = self.split_cp_arg(args['dest'])
            del args['src']
            del args['dest']

            if src_container != '':
                direction |= Direction.fromContainer.value
            if dst_container != '':
                direction |= Direction.toContainer.value

            for case in switch(direction):
                if case(Direction.fromContainer.value):
                    self.copy_from_container(
                        args, src_container, src_path, dst_path)
                    break
                if case(Direction.toContainer.value):
                    self.copy_to_container(
                        args, src_path, dst_container, dst_path)
                    break
                if case(Direction.acrossContainers.value):
                    self.settings[
                        self.name] = "copying between containers is not supported"
                    break
                if case():
                    self.settings[
                        self.name] = "must specify at least one container source"
        except (APIError, AttributeError, ValueError) as e:
            raise e

    def final(self):
        return self.settings[self.name]



"""This module contains `docker info` class"""

import json

from .command import Command


class Info(Command):
    """This class implements `docker info` command"""

    name = "info"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        self.settings[self.name] = json.dumps(
            self.client.info(), indent=4) + "\n"

    def final(self):
        return self.settings[self.name]



"""This module contains `docker container exec` class"""

from .command import Command


class Exec(Command):
    """This class implements `docker container exec` command"""

    name = "container exec"
    require = ["exec_create", "exec_start"]

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        """Execute a command in a running container"""
        self.settings[self.name] = "\r"

    def final(self):
        return self.settings[self.name]



"""This module contains `docker container run` class"""

from .command import Command


class Run(Command):
    """This class implements `docker container run` command"""

    name = "container run"
    require = ["create", "start"]

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        if args["detach"] is False and not args["tty"]:
            for line in self.client.logs(
                    self.settings["create"], stream=True):
                print line,
            self.settings[self.name] = "\r"
        else:
            self.settings[self.name] = self.settings["create"]["Id"]

    def final(self):
        return self.settings



"""This module contains base class of `docker volume command` class"""

from abc import ABCMeta, abstractmethod
from six import with_metaclass


class Command(with_metaclass(ABCMeta, object)):
    """
    This class represents a command, it must be extended
    for any class which implements a type of command
    """

    name = "command"
    client = None
    require = []
    settings = {}

    @abstractmethod
    def eval_command(self, args):
        """Evalute the command"""
        pass

    def final(self):
        """Do the final job"""
        return self.settings

    def load_require(self, args, client):
        """Load require dependencies"""
        for command in self.require:
            mod = __import__(command,
                             globals(), locals(), ['dummy'], -1)
            mod_instance = getattr(mod, command.capitalize())()
            self.settings[command] = mod_instance.command(args, client)
            del mod

    def command(self, args, client):
        """Command class entrypoint"""
        self.client = client
        self.eval_command(args)
        return self.final()



"""This module contains Logger class"""

import sys


class Logger(object):
    """Logger class"""

    # Color codes
    STD = "\033[0;0m"
    BLUE = "\033[1;34m"
    RED = "\033[0;31m"
    GREEN = "\033[0;32m"
    YELLOW = "\033[1;33m"

    def __init__(self):
        pass

    @classmethod
    def log(cls, fmt_string, *args):
        if len(args) == 0:
            print fmt_string,
        else:
            print fmt_string.format(*args),

    @classmethod
    def logInfo(cls, fmt_string, *args):
        sys.stdout.write(cls.BLUE)
        cls.log(fmt_string, *args)
        sys.stdout.write(cls.STD)

    @classmethod
    def logWarning(cls, fmt_string, *args):
        sys.stdout.write(cls.YELLOW)
        cls.log(fmt_string, *args)
        sys.stdout.write(cls.STD)

    @classmethod
    def logError(cls, fmt_string, *args):
        sys.stdout.write(cls.RED)
        cls.log(fmt_string, *args)
        sys.stdout.write(cls.STD)

    @classmethod
    def logSuccess(cls, fmt_string, *args):
        sys.stdout.write(cls.GREEN)
        cls.log(fmt_string, *args)
        sys.stdout.write(cls.STD)

    @classmethod
    def logProgressInfo(cls, fmt_string, *args):
        sys.stdout.write(cls.BLUE)
        if len(args) == 0:
            sys.stdout.write(fmt_string)
        else:
            sys.stdout.write(fmt_string.format(*args))
        sys.stdout.write(cls.STD)



"""This module contains `docker container top` class"""

import pystache
from pytabwriter import TabWriter

from .command import Command


class Top(Command):
    """This class implements `docker container top` command"""

    name = "container top"
    require = []

    defaultTemplate = '{{Pid}}\t{{User}}\t{{Time}}\t{{{Command}}}'

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        tw = TabWriter()
        tw.padding = [8, 8, 8]
        tw.writeln(
            "PID\tUSER\tTime\tCOMMAND")
        infos = self.client.top(**args)
        if infos:
            for info in infos["Processes"]:
                process = {}
                process["Pid"] = info[0]
                process["User"] = info[1]
                process["Time"] = info[2]
                process["Command"] = info[3]
                tw.writeln(pystache.render(self.defaultTemplate, process))
        self.settings[self.name] = str(tw)

    def final(self):
        return self.settings[self.name]



"""This module contains `docker volume inspect` class"""

import json
import pystache
from docker.errors import APIError

from .command import Command


class Inspect(Command):
    """This class implements `docker volume inspect` command"""

    name = "volume inspect"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        try:
            Volumes = []
            volumes = args["volumes"]
            del args["volumes"]
            for Volume in volumes:
                if args["format"]:
                    Volumes.append(pystache.render(
                        args["format"], self.client.inspect_volume(Volume)))
                else:
                    Volumes.append(json.dumps(
                        self.client.inspect_volume(Volume), indent=4))
            self.settings[self.name] = '\n'.join(Volumes)
        except APIError as e:
            raise e

    def final(self):
        return self.settings[self.name]



"""This module contains `docker network disconnect` class"""

from .command import Command


def pprint_things(l):
    """Pretty print"""
    return ''.join("{:25}".format(e) for e in l.split("\t")) + "\n"


class Disconnect(Command):
    """This class implements `docker network disconnect` command"""

    name = "network disconnect"
    require = []


    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        self.client.disconnect_container_from_network(**args)
        self.settings[self.name] = "Done!"

    def final(self):
        return self.settings[self.name]



"""This module contains Exec_start class"""

import dockerpty
from docker.errors import APIError

from .command import Command


class Exec_start(Command):
    """This class start a previously set up exec instance"""

    name = "exec_start"
    require = ["exec_create"]

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        """Start a previously set up exec instance"""
        try:
            detach = args["detach"]
            del args["detach"]
            dockerpty.start_exec(
                self.client, self.settings["exec_create"], interactive=args["interactive"])
        except APIError as e:
            raise e

    def final(self):
        return self.settings[self.name]



"""This module provides a workaround function using json.JSONDecoder.raw_decode"""

import re
import json

nonspace = re.compile(r'\S')


def json_iterparse(j):
    """A workaround function using json.JSONDecoder.raw_decode"""
    decoder = json.JSONDecoder()
    pos = 0
    while True:
        matched = nonspace.search(j, pos)
        if not matched:
            break
        pos = matched.start()
        decoded, pos = decoder.raw_decode(j, pos)
        yield decoded



"""This module contains Docker class"""

import platform
import os

from docker import APIClient

try:
    import urlparse
except ImportError:  # For Python 3
    import urllib.parse as urlparse


class Docker(object):
    """Class for manipulating the docker client."""

    __stack = None

    host = None
    client = None
    ctr = None
    current_ctr = None
    buf = None

    category = None
    command_flag = None

    def __init__(self, host='127.0.0.1'):
        """Loading docker environments"""
        if platform.system() == 'Darwin' or platform.system() == 'Windows':
            try:
                # TLS problem, can be referenced from
                # https://github.com/docker/machine/issues/1335
                from docker.utils import kwargs_from_env
                self.host = '{0}'.format(urlparse.urlparse(
                    os.environ['DOCKER_HOST']).netloc.split(':')[0])
                self.client = APIClient(
                    base_url='{0}'.format(os.environ['DOCKER_HOST']))
                kwargs = kwargs_from_env()
                kwargs['tls'].assert_hostname = False
                self.client = APIClient(**kwargs)
            except KeyError:
                self.host = host
                self.client = APIClient(base_url='unix://var/run/docker.sock')
        else:
            self.host = host
            self.client = APIClient(base_url='unix://var/run/docker.sock')
        self.client.ping()

    def __dry(self):
        """Load the command and configure the environment with dry-run"""
        self.buf = "dry-run complete"

    def load(self, args, dry=False):
        """Load the command_flag and configure the environment"""
        if dry:
            self.__dry()
        else:
            self.category = args["manage_flag"]
            del args["manage_flag"]
            if args.get("{}_flag".format(self.category)):
                self.command_flag = args["{}_flag".format(self.category)]
                del args["{}_flag".format(self.category)]
            else:
                self.command_flag = self.category
                self.category = 'system'
            mod = __import__("Docker.{}.{}".format(self.category.capitalize(), self.command_flag),
                             globals(), locals(), ['dummy'], -1)
            self.__intrude(mod)
            mod_instance = getattr(mod, self.command_flag.capitalize())()

            if mod_instance.require:
                mod_instance.load_require(args, self.client)

            self.buf = mod_instance.command(args, self.client)

    def buffer(self):
        """Receive outcome"""
        return self.buf

    def set_buffer(self, buf):
        """Set received outcome"""
        self.buf = buf

    def push(self, **kwargs):
        """Push custom classes into Docker.__stack"""
        self.__stack = kwargs

    def __intrude(self, mod):
        """Intrude classes"""
        if self.__stack:
            for k, v in self.__stack.iteritems():
                override, value = k.split('|')
                if "tsaotun.lib.Docker." + override == mod.__name__:
                    setattr(mod, value, v)



"""This module contains `docker image pull` class"""

import colorama

from .command import Command
from ...Utils.json_iterparse import json_iterparse


colorama.init()
info = {
    "title": "",
    "layers": {},
    "final": []
}
MINY, OFFSET = 1, 2


def put_cursor(x, y):
    print "\x1b[{};{}H".format(y + 1, x + 1)


def clear():
    print "\x1b[2J"


class Pull(Command):
    """This class implements `docker image pull` command"""

    name = "image pull"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def preprocess(self, args):
        try:
            repository, tag = args["image"][0].split(':', 2)
        except ValueError:
            repository, tag = args["image"][0], None
        del args["image"]
        args["repository"] = repository
        if tag:
            args["tag"] = tag
        else:
            args["tag"] = "latest"
        args["stream"] = True

    def output(self, infos, args):
        try:
            print infos
            if infos["id"] == args["tag"]:
                info["title"] = "{}: {}".format(infos["id"], infos["status"])
            elif "Pulling" in infos["status"]:
                info["layers"][infos["id"]] = {"status": infos[
                    "status"], "progressDetail": infos["progressDetail"]}
            else:
                for infos_key in infos.keys():
                    for info_key in info["layers"][infos["id"]]:
                        if infos_key == info_key:
                            info["layers"][infos["id"]][
                                info_key] = infos[infos_key]
                            del infos[infos_key]
                for infos_key in infos.keys():
                    if infos_key != "id":
                        info["layers"][infos["id"]][
                            infos_key] = infos[infos_key]
                clear()
                put_cursor(0, MINY)

                tmp = []
                for info_key, info_value in info["layers"].iteritems():
                    if "progress" in info_value:
                        tmp.append("%s: %s: %s" % (info_key, info_value[
                            "status"], info_value["progress"]))
                    else:
                        tmp.append("%s: %s: %s" % (info_key, info_value[
                            "status"], info_value["progressDetail"]))

                s = '\n'.join(tmp)
                line_n = len(s.split('\n'))
                print "%s\n%s" % (info["title"], s)
                return line_n
        except KeyError:
            if len(info["final"]):
                info["final"].append(infos["status"])
                clear()
                put_cursor(0, MINY)
                s = '\n'.join(info["final"])
                line_n = len(s.split('\n'))
                print "%s\n%s" % (info["title"], s)
                return line_n
            else:
                info["final"].append(infos["status"])

    def eval_command(self, args):
        self.preprocess(args)
        line_n = 0
        try:
            for line in self.client.pull(**args):
                for iterElement in list(json_iterparse(line)):
                    line_n = self.output(iterElement, args)
        except KeyboardInterrupt:
            put_cursor(0, MINY + OFFSET + line_n)
            colorama.deinit()
            raise KeyboardInterrupt
        put_cursor(0, MINY + OFFSET + line_n)
        colorama.deinit()
        self.settings[self.name] = "\r"

    def final(self):
        return self.settings[self.name]



"""This module contains Exec_create class"""

import dockerpty
from docker.errors import APIError

from .command import Command


class Exec_create(Command):
    """This class sets up an exec instance in a running container"""

    name = "exec_create"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        """Set up an exec instance in a running container"""
        try:
            detach = args["detach"]
            del args["detach"]
            self.settings[self.name] = dockerpty.exec_create(self.client, args["container"], args["cmd"], interactive=args["interactive"])
            args["detach"] = detach
        except APIError as e:
            raise e

    def final(self):
        return self.settings[self.name]



"""This module contains `docker image images` class"""

import arrow
import humanize
import pystache
from pytabwriter import TabWriter

from .command import Command


class Images(Command):
    """This class implements `docker image images` command"""

    name = "image images"
    require = []

    defaultTemplate = '{{{Repository}}}\t{{{Tag}}}\t{{Id}}\t{{Created}}\t{{Size}}'
    digestsTemplate = '{{{Repository}}}\t{{{Tag}}}\t{{{Digest}}}\t{{Id}}\t{{Created}}\t{{Size}}'

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        tw = TabWriter()
        if args["digests"]:
            tw.padding = [3, 10, 3, 8, 8]
            fm = self.digestsTemplate
            tw.writeln(
                "REPOSITORY\tTAG\tDIGEST\tIMAGE ID\tCREATED\tSIZE")
        elif args["format"] is None:
            tw.padding = [3, 10, 8, 8]
            fm = self.defaultTemplate
            tw.writeln(
                "REPOSITORY\tTAG\tIMAGE ID\tCREATED\tSIZE")
        else:
            fm = args["format"]
            self.settings[self.name] = ""
        del args["digests"]
        del args["format"]

        args["filters"] = dict(args["filters"]) if args["filters"] else None

        nodes = self.client.images(**args)
        for node in nodes:
            try:
                node["Repository"], node["Tag"] = node[
                    "RepoTags"][0].split(":")
            except TypeError:
                node["Repository"] = node["RepoDigests"][0].split('@', 2)[0]
                node["Tag"] = "<none>"
            node["Digest"] = node["RepoDigests"][0].split('@', 2)[1] if node[
                "RepoDigests"] else '<' + str(node["RepoDigests"]) + '>'
            node["Id"] = node["Id"].split(":")[1][:12]
            node["Created"] = arrow.get(node["Created"]).humanize()
            node["Size"] = humanize.naturalsize(node["VirtualSize"])
            tw.writeln(pystache.render(fm, node))
        self.settings[self.name] = str(tw)

    def final(self):
        return self.settings[self.name]



"""This module contains `docker inspect` class"""

import json
import pystache
from docker.errors import APIError

from .command import Command


class Inspect(Command):
    """This class implements `docker inspect` command"""

    name = "inspect"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        try:
            obj = args["object"]
            del args["object"]
            try:
                if args["format"]:
                    self.settings[self.name] = pystache.render(args["format"], self.client.inspect_container(obj))
                else:
                    self.settings[self.name] = json.dumps(
                        self.client.inspect_container(obj), indent=4) + "\n"
            except APIError as e:
                if "No such container" in str(e):
                    try:
                        if args["format"]:
                            self.settings[self.name] = pystache.render(args["format"], self.client.inspect_image(obj))
                        else:
                            self.settings[self.name] = json.dumps(
                                self.client.inspect_image(obj), indent=4) + "\n"
                    except APIError as ev:
                        raise ev
                else:
                    raise e
        except APIError as e:
            raise e

    def final(self):
        return self.settings[self.name]



"""This module contains `docker image save` class"""

from .command import Command


class Save(Command):
    """This class implements `docker image save` command"""

    name = "image save"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        image = self.client.get_image(args["image"])
        if args["output"]:
            with open(args["output"], 'w') as f:
                f.write(image.data)
            self.settings[self.name] = "\r"
        else:
            self.settings[self.name] = image.data

    def final(self):
        return self.settings[self.name]



# -*- coding: UTF-8 -*-
"""
This module provides converting between datetime.timedelta and Go's Duration strings.
Taken from https://github.com/icholy/Duration.py/, and made by icholy.
"""

import re
import datetime

_nanosecond_size = 1
_microsecond_size = 1000 * _nanosecond_size
_millisecond_size = 1000 * _microsecond_size
_second_size = 1000 * _millisecond_size
_minute_size = 60 * _second_size
_hour_size = 60 * _minute_size

units = {
    "ns": _nanosecond_size,
    "us": _microsecond_size,
    "µs": _microsecond_size,
    "μs": _microsecond_size,
    "ms": _millisecond_size,
    "s": _second_size,
    "m": _minute_size,
    "h": _hour_size
}


def from_str(duration):
    """Parse a duration string to a datetime.timedelta"""

    if duration in ("0", "+0", "-0"):
        return datetime.timedelta()

    pattern = re.compile('([\-\+\d\.]+)([a-zµμ]+)')
    total = 0
    sign = -1 if duration[0] == '-' else 1
    matches = pattern.findall(duration)

    if not len(matches):
        raise Exception("Invalid duration {}".format(duration))

    for (value, unit) in matches:
        if unit not in units:
            raise Exception(
                "Unknown unit {} in duration {}".format(unit, duration))
        try:
            total += float(value) * units[unit]
        except:
            raise Exception(
                "Invalid value {} in duration {}".format(value, duration))

    microseconds = total / _microsecond_size
    return datetime.timedelta(microseconds=sign * microseconds)


def to_str(delta):
    """Format a datetime.timedelta to a duration string"""

    total_seconds = delta.total_seconds()
    sign = "-" if total_seconds < 0 else ""
    nanoseconds = abs(total_seconds * _second_size)

    if total_seconds < 1:
        result_str = _to_str_small(nanoseconds)
    else:
        result_str = _to_str_large(nanoseconds)

    return "{}{}".format(sign, result_str)


def _to_str_small(nanoseconds):

    result_str = ""

    if not nanoseconds:
        return "0"

    milliseconds = int(nanoseconds / _millisecond_size)
    if milliseconds:
        nanoseconds -= _millisecond_size * milliseconds
        result_str += "{}ms".format(milliseconds)

    microseconds = int(nanoseconds / _microsecond_size)
    if microseconds:
        nanoseconds -= _microsecond_size * microseconds
        result_str += "{}us".format(microseconds)

    if nanoseconds:
        result_str += "{}ns".format(nanoseconds)

    return result_str


def _to_str_large(nanoseconds):

    result_str = ""

    hours = int(nanoseconds / _hour_size)
    if hours:
        nanoseconds -= _hour_size * hours
        result_str += "{}h".format(hours)

    minutes = int(nanoseconds / _minute_size)
    if minutes:
        nanoseconds -= _minute_size * minutes
        result_str += "{}m".format(minutes)

    seconds = float(nanoseconds) / float(_second_size)
    if seconds:
        nanoseconds -= _second_size * seconds
        result_str += "{}s".format(seconds)

    return result_str



"""This module contains `docker image history` class"""

import arrow
import humanize
import pystache
from pytabwriter import TabWriter

from .command import Command


class History(Command):
    """This class implements `docker image history` command"""

    name = "image history"
    require = []

    defaultTemplate = '{{{Id}}}@{{Created}}@{{{CreatedBy}}}@{{Size}}@{{Comment}}'

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        tw = TabWriter(tabchar='@')
        tw.padding = [8, 9, 3, 8]
        fm = self.defaultTemplate
        tw.writeln(
            "IMAGE@CREATED@CREATED BY@SIZE@COMMENT")
        nodes = self.client.history(**args)
        if nodes:
            for node in nodes:
                node["Id"] = node["Id"].split(":")[1][:12] if ":" in node[
                    "Id"] else node["Id"]
                node["Created"] = arrow.get(node["Created"]).humanize()
                node["CreatedBy"] = (node["CreatedBy"][
                    :42] + "...") if len(node["CreatedBy"]) >= 45 else node["CreatedBy"]
                node["Size"] = humanize.naturalsize(node["Size"])
                tw.writeln(pystache.render(fm, node))
        self.settings[self.name] = str(tw)

    def final(self):
        return self.settings[self.name]



"""This module contains `docker container stats` class"""

import colorama
import humanize
import pystache

from .command import Command


colorama.init()
info = {
    "title": "",
    "layers": {},
    "final": []
}


def pprint_things(l):
    """Pretty print"""
    return ''.join("{:20}".format(e) for e in l.split("\t")) + "\n"


def put_cursor(x, y):
    print "\x1b[{};{}H".format(y + 1, x + 1)

def clear():
    print "\x1b[2J"


class Stats(Command):
    """This class implements `docker container stats` command"""

    name = "container stats"
    require = []

    defaultTemplate = '{{Id}}\t{{Cpu}}%\t{{MemUsage}} / {{Limit}}\t{{Mem}}%\t{{NetInput}} / {{NetOutput}}\t0 B / 0 B\t{{Pids}}'

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        try:
            stats = []
            containers = args["containers"]
            del args["containers"]
            args["decode"] = True
            for container in containers:
                args["container"] = container
                stats.append(self.client.stats(**args))
                """
                for line in self.client.stats(**args):
                    for iterElement in list(json_iterparse.json_iterparse(line)):
                        # self.output(iterElement, args)
                        print iterElement
                """
            clear()
            put_cursor(0, 0)
            print pprint_things(
                "CONTAINER\tCPU %\tMEM USAGE / LIMIT\tMEM %\tNET I/O\tBLOCK I/O\tPIDS"),
            while True:
                y = 1
                for stat in stats:
                    put_cursor(0, y)
                    y += 1
                    tmp = next(stat)
                    tmp["Id"] = tmp["id"][:12]
                    tmp["Cpu"] = (tmp["cpu_stats"]["cpu_usage"][
                        "total_usage"] / tmp["cpu_stats"]["system_cpu_usage"])
                    tmp["MemUsage"] = humanize.naturalsize(
                        tmp["memory_stats"]["usage"])
                    tmp["Limit"] = humanize.naturalsize(
                        tmp["memory_stats"]["limit"])
                    tmp["Mem"] = (tmp["memory_stats"]["usage"] /
                                  tmp["memory_stats"]["limit"])
                    tmp["NetInput"] = humanize.naturalsize(
                        tmp["networks"]["eth0"]["rx_bytes"])
                    tmp["NetOutput"] = humanize.naturalsize(
                        tmp["networks"]["eth0"]["tx_bytes"])
                    tmp["Pids"] = tmp["pids_stats"]["current"]
                    print pprint_things(pystache.render(self.defaultTemplate, tmp))
        except KeyboardInterrupt:
            put_cursor(0, y)
            colorama.deinit()
            raise KeyboardInterrupt
        put_cursor(0, y)
        colorama.deinit()
        self.settings[self.name] = "\r"

    def final(self):
        return self.settings[self.name]



"""This module contains `docker network create` class"""


from docker.types import (IPAMConfig, IPAMPool)
from docker.errors import APIError

from .command import Command


class Create(Command):
    """This class implements `docker network create` command"""

    name = "network create"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def preprocess(self, args):
        """Preprocess arguments provided for `docker network create`"""
        pool = {}

        # subnet
        pool["subnet"] = args["subnet"]
        del args["subnet"]

        # iprange
        pool["iprange"] = args["iprange"]
        del args["iprange"]

        # gateway
        pool["gateway"] = args["gateway"]
        del args["gateway"]

        # create ipam
        ipam_pool = IPAMPool(**pool)
        ipam_config = IPAMConfig(
            driver=args["ipamdriver"],
            pool_configs=[ipam_pool],
            options=dict(args["ipamopt"]) if args["ipamopt"] else None
        )
        del args["ipamdriver"]
        del args["ipamopt"]
        if "None" in str(ipam_config):
            return None
        else:
            return ipam_config

    def eval_command(self, args):
        """Create host config for containers"""
        try:
            args["ipam"] = self.preprocess(args)
            args["labels"] = dict(args["labels"]) if args["labels"] else None
            args["options"] = dict(args["options"]) if args["options"] else None
            self.settings[self.name] = self.client.create_network(**args)
        except APIError as e:
            raise e

    def final(self):
        return self.settings[self.name]



"""This module contains `docker network ls` class"""

from collections import defaultdict
import pystache
from pytabwriter import TabWriter

from .command import Command


class Ls(Command):
    """This class implements `docker network ls` command"""

    name = "network ls"
    require = []

    defaultTemplate = '{{Id}}\t{{Name}}\t{{Driver}}\t{{Scope}}'

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        tw = TabWriter()
        if args["format"] is None:
            tw.padding = [8, 14, 14]
            fm = self.defaultTemplate
            tw.writeln(
                "NETWORK ID\tNAME\tDRIVER\tSCOPE")
        else:
            fm = args["format"]
        del args["format"]

        if args["filters"]:
            filters = args["filters"]
            args["filters"] = []
            d = defaultdict(list)
            for k, v in filters:
                d[k].append(v)
            args["filters"] = dict(d)

        # wait for PR: https://github.com/docker/docker-py/pull/1362
        del args["filters"]

        nodes = self.client.networks(**args)
        if nodes:
            for node in nodes:
                node["Id"] = node["Id"][:12]
                tw.writeln(pystache.render(fm, node))
        self.settings[self.name] = str(tw)

    def final(self):
        return self.settings[self.name]



"""This module contains `docker network inspect` class"""

import json
import pystache
from docker.errors import APIError

from .command import Command


class Inspect(Command):
    """This class implements `docker network inspect` command"""

    name = "network inspect"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        try:
            Networks = []
            networks = args["networks"]
            del args["networks"]
            for Network in networks:
                if args["format"]:
                    Networks.append(pystache.render(
                        args["format"], self.client.inspect_network(Network)))
                else:
                    Networks.append(json.dumps(
                        self.client.inspect_network(Network), indent=4))
            self.settings[self.name] = '\n'.join(Networks)
        except APIError as e:
            raise e

    def final(self):
        return self.settings[self.name]



"""This module contains `docker network remove` class"""

from docker.errors import APIError

from .command import Command


class Remove(Command):
    """This class implements `docker network remove` command"""

    name = "network remove"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        try:
            Ids = []
            networks = args["networks"]
            del args["networks"]
            for Id in networks:
                Ids.append(Id)
                self.client.remove_network(Id)
            self.settings[self.name] = '\n'.join(Ids)
        except APIError as e:
            raise e

    def final(self):
        return self.settings[self.name]



"""python setup script"""
import codecs
import os
import re

from setuptools import setup, find_packages

ROOT_DIR = os.path.dirname(__file__)
SOURCE_DIR = os.path.join(ROOT_DIR)


def find_version(*file_paths):
    """
    Read the version number from a source file.
    Why read it, and not import?
    see https://groups.google.com/d/topic/pypa-dev/0PkjVpcxTzQ/discussion
    """
    # Open in Latin-1 so that we avoid encoding errors.
    # Use codecs.open for Python 2 compatibility
    with codecs.open(os.path.join(ROOT_DIR, *file_paths), 'r', 'latin1') as f:
        version_file = f.read()

    # The version line must have the form
    # __version__ = 'ver'
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string')

requirements = [
    'docker >= 2.0.1',
    'dockerpty >= 0.4.1',
    'arrow >= 0.10.0',
    'humanize >= 0.5.1',
    'pystache >= 0.5.4',
    'colorama >= 0.3.7',
    'argcomplete >= 1.7.0'
]

# Get the long description from the relevant file
with codecs.open(os.path.join(ROOT_DIR, 'README.md'), 'r', 'latin1') as f:
    long_description = f.read()

setup(name='tsaotun',
      version=find_version('tsaotun', '__init__.py'),
      description='Python based Assistance for Docker',
      long_description=long_description,
      author='Boik Su',
      author_email='boik@tdohacker.org',
      url='https://github.com/qazbnm456/tsaotun',
      download_url='https://codeload.github.com/qazbnm456/tsaotun/tar.gz/0.8.1',
      keywords=['0.8.1'],
      packages=find_packages(),
      install_requires=requirements,
      dependency_links=[
          'https://github.com/qazbnm456/pytabwriter/tarball/master#egg=pytabwriter-0.1.1'],
      entry_points="""
            [console_scripts]
            tsaotun=tsaotun.cli:cli
      """,
      classifiers=[
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: Apache Software License',
          'Topic :: Software Development',
          'Topic :: Terminals',
          'Topic :: Utilities',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7'
      ]
      )



"""This module contains `docker image remove` class"""

from docker.errors import APIError

from .command import Command


class Remove(Command):
    """This class implements `docker image remove` command"""

    name = "image remove"
    require = []

    def __init__(self):
        Command.__init__(self)
        self.settings[self.name] = None

    def eval_command(self, args):
        try:
            Images = []
            images = args['images']
            del args['images']
            for Image in images:
                Images.append(Image)
                args['image'] = Image
                self.client.remove_image(**args)
                del args['image']
            self.settings[self.name] = '\n'.join(Images)
        except APIError as e:
            raise e

    def final(self):
        return self.settings[self.name]



# -*- coding: UTF-8 -*-
"""
This module provides a recursive way to dig down through nested objects.
Taken from http://pingfive.typepad.com/blog/2010/04/deep-getattr-python-function.html, and made by Andew.
"""


def deepgetattr(obj, attr):
    """Recurses through an attribute chain to get the ultimate value"""
    return reduce(getattr, attr.split('.'), obj)
